#include "music.h"
#include "mididev.h"

void music::stream::stack::push(unsigned pos, unsigned kertoja)
{
    item i(pos, kertoja);
    pino.push_back(i);
}

unsigned music::stream::stack::pop(unsigned nextpos)
{
    unsigned c = size();
    if(!c)return nextpos;
    unsigned n = pino[--c].kertoja;
    unsigned p = pino[c].alku;
    if(n)
        if(!(pino[c].kertoja = n-1))
        {
            pino.erase(pino.end() - 1);
            return nextpos;
        }
    return p;
}

void music::stream::cachedctrl(int ctrl, int value)
{
    if(owner->streamscache[curchan*128+ctrl] != value)
    {
        owner->streamscache[curchan*128+ctrl] = value;
        if(ctrl == -1) // special
            owner->driver->patch(curchan, instru);
        else
            owner->driver->ctrl(curchan, ctrl, value);
    }
}

void music::stream::updstat()
{
    if(curchan < 0)return;
    
    // use cachedctrl to decrease the amount of traffic.    
    cachedctrl(0, bank);
    cachedctrl(-1, instru); //hack
    cachedctrl(91, reverb*127/99);
    cachedctrl(93, chorus*127/99);
    if(curchan != 9)
        cachedctrl(7, volume*127/99);
    cachedctrl(10, pan*127/99);
}

void music::stream::endnote()
{
    if(noteoffchan >= 0)
    {
        owner->driver->noteoff(noteoffchan, noteoffnot, 0);
        noteoffchan = -1;
        if(curchan >= 0)
        {
            owner->freechan(curchan);
            curchan = -1;
        }
    }
}

void music::stream::startnote(int notenum)
{
    if(instru < 0)
    {
        curchan = 9;
        updstat();
        owner->driver->noteon(noteoffchan=9, noteoffnot=-instru, volume*velocity/99);
        curchan = -1;
    }
    else
    {
        curchan = owner->allochan();
        updstat();
        owner->driver->noteon(noteoffchan=curchan, noteoffnot=notenum, velocity*127/99);
    }
}


unsigned music::stream::readnum(int &dest, unsigned n) const
{
    if(n >= data.size() || data[n] < '0' || data[n] > '9')return 0;
    int value;
    for(value = 0; ; )
    {
        value = value * 10 + data[n] - '0';
        ++n;
        if(n >= data.size() || data[n] < '0' || data[n] > '9')break;
    }
    dest = value;
    return n;
}

bool music::stream::dead()
{
    if(wait || substreams.size())return false;
    return (tilanne == died || posi >= data.size());
}

void music::stream::die()
{
    stack.clear();
    posi = data.size();
    wait = 0;
    if(substreams.size() == 1) // This is probably 'x' case.
    {
        owner->unregisterstream(&*substreams.begin());
    }
    substreams.clear();
    endnote();
    tilanne = died;
}

bool music::stream::tick()
{
Redo:
    if(substreams.size())
    {
        unsigned n=0;
        list<stream>::iterator i;
        for(i = substreams.begin(); i != substreams.end(); ++i)
        {
            if(!i->tick())
                ++n;
        }
        if(n) return false;
        if(substreams.size() == 1) // This is probably 'x' case.
        {
            owner->unregisterstream(&*substreams.begin());
        }
        substreams.clear();
        tilanne = fetch;                        
    }
    if(owner->tickpos % divider)return false;
    if(wait > 0 && --wait)
    {
Return:
        updstat();
        return dead();
    }
    if(tilanne == nuotti)
    {
        endnote();
        tilanne = fetch;
    }
    if(dead())return true;
    if(tilanne == fetch)
        while(!wait && !dead())
        {
        //  fprintf(stderr, "I am \"%s\", executing command: '%c' at posi %u\n",
        //        data.c_str(), data[posi], posi);
        //  fflush(stderr);
            switch(data[posi])
            {
                case '/':
                    if(!(posi = readnum(divider, posi+1)))die();
                    break;
                case 'p':
                case 'y':
                {
                    bool sammu = data[posi] == 'p';
                    unsigned tmp = readnum(wait, ++posi);
                    if(tmp)
                        posi = tmp;
                    else
                        wait = length;
                    if(!wait)wait = 1;
                    wait *= divider;
                    if(sammu)endnote();
                    break;
                }
                case 'l':
                    if(!(posi = readnum(length, posi+1)))die();
                    break;
                case 'o':
                    if(!(posi = readnum(octave, posi+1)))die();
                    break;
                case '<':
                    if(octave > 0)--octave;
                    ++posi;
                    break;
                case '>':
                    if(octave < 7)++octave;
                    ++posi;
                    break;
                case 'a':
                case 'b':
                case 'c':
                case 'd':
                case 'e':
                case 'f':
                case 'g':
                case 'n':
                {
                    int notenum;
                    
                    if(data[posi] == 'n')
                    {
                        if(!(posi = readnum(notenum, posi+1)))die();
                        if(data[posi]==':')++posi;
                    }
                    else
                    {
                        static const char n[7] = {9,11,0,2,4,5,7};
                        notenum = n[data[posi++]-'a'] + octave*12;
                        for(; !dead() && (data[posi] == '-' || data[posi]=='+'); ++posi)
                            if(data[posi] == '-') --notenum; else ++notenum;
                    }
                    
                    endnote();
                    startnote(notenum);
                    
                    unsigned tmp = readnum(wait, posi);
                    if(tmp)
                        posi = tmp;
                    else
                        wait = length;
                    if(!wait)wait = 1;
                    wait *= divider;
                    if(data[posi]=='.') ++posi; else tilanne = nuotti;
                    break;
                }
                case 'v':
                    if(!(posi = readnum(volume, posi+1)))die();
                    break;
                case 'h':
                    if(!(posi = readnum(chorus, posi+1)))die();
                    break;
                case 'r':
                    if(!(posi = readnum(reverb, posi+1)))die();
                    break;
                case 'm':
                    if(!(posi = readnum(pan, posi+1)))die();
                    break;
                case 's':
                    if(!(posi = readnum(velocity, posi+1)))die();
                    break;
                case 't':
                {
                    unsigned tmp;
                    int a, b;
                    if(!(tmp = readnum(a, posi+1)))die();
                    if(data[tmp] != ':')die();
                    if(!(tmp = readnum(b, ++tmp)))die();
                    
                    if(b && (owner->tickpos / divider) % b == (unsigned)a)posi = tmp;
                    goto Return;
                }
                case 'w':
                {
                    bool accept=false;
                    unsigned tmp=posi;
                    do {
                        int num;
                        if(!(tmp = readnum(num, tmp+1)))die();
                        if(!accept && owner->isplaying(num))accept = true;
                    } while(data[tmp] == ',');
                    if(accept)posi=tmp;
                    goto Return;
                }
                case '(':
                {
                    int num;
                    if(!(posi = readnum(num, posi+1)))die();
                    unsigned a, b;
                    for(a=posi, b=0; posi < data.size(); ++b, ++posi)
                        if(data[posi] == ')')
                            break;
                    if(posi >= data.size())
                        die();
                    else
                    {
                        ++posi;
                    //  fprintf(stderr, "Defining subproc number %d as '%s'\n", num, data.substr(a,b).c_str());
                    //  fflush(stderr);
                        owner->installstream(num, data.substr(a, b));
                    }
                    break;
                }
                case '{':
                    ++posi;
                    do {
                        unsigned a, b;
                        for(a=posi, b=0; posi < data.size(); ++b, ++posi)
                            if(data[posi] == ',' || data[posi] == '}')
                                break;
                        if(posi >= data.size())
                        {
                            die();
                            break;
                        }
                        stream tmp(data.substr(a, b));
                        tmp.setowner(owner);
                        tmp.restart(*this);
                        substreams.push_back(tmp);
                    } while(data[posi++] == ',');
                    goto Redo;
                case '[':
                {
                    int kertoja = 0;
                    unsigned tmp = readnum(kertoja, ++posi);
                    if(tmp)posi = tmp;
                    stack.push(posi, kertoja);
                    break;
                }
                case ']':
                {
                    posi = stack.pop(posi+1);
                    break;
                }
                case 'x':
                {
                    int num;
                    if(!(posi = readnum(num, posi+1)))
                        die();
                    else
                    {
                        stream tmp(owner->program[num]);
                        tmp.setowner(owner);
                        tmp.restart(*this);
                        substreams.push_back(tmp);
                        owner->registerstream(&*substreams.begin(), num);
                    }
                    goto Redo;
                }
                case 'k':
                {
                    int tmp;
                    unsigned tmp2 = readnum(tmp, ++posi);
                    if(tmp2)
                    {
                        posi = tmp2;
                        owner->killallstreamsnamed(tmp);
                    }
                    else
                        owner->killallstreams();
                    break;
                }
                case 'i':
                {
                    bool miinus = false;
                    if(data[++posi] == '-') { miinus = true; ++posi; }
                    int tmp;
                    if(!(posi = readnum(tmp, posi)))die();
                    if(miinus)tmp = -tmp;
                    instru = tmp;
                    if(data[posi] != ':')
                        bank = 0;
                    else
                        if(!(posi = readnum(bank, ++posi)))die();
                    updstat();
                    break;
                }
                case ' ':
                case '\t':
                case '\r':
                case '\n':
                    ++posi;
                    break;
                case '\33':
                    if(data[++posi] != '['
                    || data[++posi] != '2'
                    || data[++posi] != 'c')
                        die();
                    else
                        ++posi;
                    break;
                default:
                    die();
            }
        }
    goto Return;
}

void music::tick()
{
    list<stream>::iterator i;
    for(i=streams.begin(); i!=streams.end(); )
    {
        i->tick();
        if(i->dead())
        {
            list<stream>::iterator j = i;
            ++j;
            streams.erase(i);
            i = j;
        }
        else
            ++i;
    }
    ++tickpos;
//    if(!(tickpos % 100))driver->sleep(1);
}

void music::upload(const string &s)
{
    if(s[0] == '?')
    {
        string code, arg;
        unsigned p = 1;
        
        while(p < s.size() && isalpha(s[p]))
            code += s[p++];
        
        arg = s.substr(p);
        
        if(code == "list")
        {
            fprintf(stderr, "\n---- Defined programs ----\n");
            map<int, string>::iterator i;
            
            for(i = program.begin(); i != program.end(); ++i)
            {
                fprintf(stderr, "%5d: %s", i->first, i->second.c_str());
                
                map<stream *, int>::iterator j;
                unsigned c=0;
                for(j = registeredstreams.begin(); j != registeredstreams.end(); ++j)
                    if(j->second == i->first)
                        c++;
                
                if(c)
                {
                    fprintf(stderr, " (Running");
                    if(c > 1)fprintf(stderr, " %u times", c);
                    fprintf(stderr, ")");
                }
                
                fprintf(stderr, "\n");
            }
        }
        else if(code == "wipe")
        {
            list<stream>::iterator i;
            for(i=streams.begin(); i!=streams.end(); ++i)i->die();
            fprintf(stderr, "\n--- All running streams were killed.");
            program.clear();
            fprintf(stderr, "\n--- All programmed timbres were deleted.\n");
        }
        else
            fprintf(stderr, "\n--- Invalid command. Available: ?list, ?wipe ----\n");
        
        return;
    }
    stream tmp(s);
    tmp.setowner(this);
    streams.push_back(tmp);
}
