#ifndef pokemud_music_h
#define pokemud_music_h

#include <string>
#include <vector>
#include <list>
#include <set>
#include <map>

/* noteoff(ch, note, pressure)
   noteon(ch, note, pressure)
   ctrl(ch, ctrl, value)
   patch(ch, ins)
*/

class music
{
public:
    class mididev *driver;
    unsigned tickpos;
    
    map<unsigned, int> streamscache;
    
private:
    class stream
    {
    private:
        class stack
        {
        public:
            class item
            {
            public:
                unsigned alku, kertoja;
                item(unsigned posi, unsigned n) : alku(posi), kertoja(n) { }
            };
            stack() { }
            ~stack() { }
            void clear() { pino.clear(); }
            unsigned size() const { return pino.size(); }
            void push(unsigned pos, unsigned kertoja);
            unsigned pop(unsigned nextpos);
        private:
            vector<item> pino;
        } stack;
    private:
        class music *owner;
        string data;
        unsigned posi;
        enum { fetch, nuotti, died } tilanne;
        int wait;
        int divider, length, octave, volume;
        int instru, bank, chorus, reverb, pan;
        int velocity;
        list<stream> substreams;
        int curchan;
        int noteoffchan, noteoffnot;
        
        // returns modified posi (pointing after number) (0 if there was no number);
        unsigned readnum(int &dest, unsigned n) const;
        
        void cachedctrl(int ctrl, int value);        
        void updstat();
        void endnote();
        void startnote(int notenum);
        
    public:
        stream() : noteoffchan(-1)
        {
            tilanne = died;
        }
        stream(const string &s) : noteoffchan(-1)
        {
            install(s);
        }
        void install(const string &s)
        {
            data = s;
            restart();
        }
        ~stream()
        {
            if(substreams.size() == 1) // This is probably 'x' case.
            {
                owner->unregisterstream(&*substreams.begin());
            }
            endnote();
        }
        void setowner(class music *o)
        {
            owner = o;
        }
        void die();
        bool dead();
        
        // gives default attributes
        void restart()
        {
            posi = 0;
            wait = 0;
            tilanne = fetch;
            divider = 4;
            length  = 4;
            octave  = 5;
            volume  = 80;
            instru  = 1;
            bank    = 0;
            chorus  = 25;
            reverb  = 50;
            pan     = 50;
            velocity= 80;
            curchan = -1;
            noteoffchan = -1;
        }
        // gives attributes from parent.
        void restart(const stream &parent)
        {
            restart();
            divider = parent.divider;
            length  = parent.length;
            octave  = parent.octave;
            volume  = parent.volume;
            instru  = parent.instru;        
            bank    = parent.bank;
            chorus  = parent.chorus;
            reverb  = parent.reverb;
            pan     = parent.pan;
            velocity= parent.velocity;
        }
        // returns true if finished, false if still something to do.
        bool tick();
    };
private:
    bool chans[15];
    list<stream> streams;
    map<stream *, int> registeredstreams;
public:
    map<int, string> program;
public:
    int allochan()
    {
        unsigned n;
        for(n=0; n<15; ++n)
            if(!chans[n])
            {
                chans[n] = true;
                return n;
            }
        return 0;
    }
    void freechan(int n)
    {
        if(n > 9)--n;
        if(n > 0 && n < 15)chans[n] = false;
    }
    music() : driver(NULL), tickpos(0)
    {
        for(unsigned a=0; a<15; chans[a++] = false);
    }
    music(class mididev *drv) : driver(drv), tickpos(0)
    {
        for(unsigned a=0; a<15; chans[a++] = false);
    }
    ~music()
    {
    }
    void registerstream(stream *str, int num)
    {
    	registeredstreams[str] = num;
    }
    void unregisterstream(stream *str)
    {
    	map<stream *, int>::iterator i;
    	i = registeredstreams.find(str);
    	if(i != registeredstreams.end())
    		registeredstreams.erase(i);
    }
    void killallstreams()
    {
    	map<stream *, int>::iterator i;
    	for(i = registeredstreams.begin(); i != registeredstreams.end(); ++i)
   			i->first->die();    	
    }
    void killallstreamsnamed(int num)
    {
    	map<stream *, int>::iterator i;
    	for(i = registeredstreams.begin(); i != registeredstreams.end(); ++i)
    		if(i->second == num)
    			i->first->die();
    }
    bool isplaying(int num)
    {
    	map<stream *, int>::iterator i;
    	for(i = registeredstreams.begin(); i != registeredstreams.end(); ++i)
    		if(i->second == num)
    			return true;
    	return false;
    }
    void installstream(int num, const string &s)
    {
        program[num] = s;
    }
    void setdriver(class mididev *drv)
    {
        driver = drv;
    }
    void upload(const string &s);
    void tick();
};

/* /WWW/midistream-kuvaus-scratch.txt */

#endif
